from typing import List

from matplotlib import figure as matplotlib_figure
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from nfoursid.state_space import StateSpace
from nfoursid.utils import Utils


class Kalman:
    r"""
    Implementation [1] of a Kalman filter for a state-space model ``state_space``:

    .. math::
        \begin{cases}
            x_{k+1} &= A x_k + B u_k + w_k \\
            y_k &= C x_k + D u_k + v_k
        \end{cases}

    The matrices :math:`(A, B, C, D)` are taken from the state-space model ``state_space``.
    The measurement-noise :math:`v_k` and process-noise :math:`w_k` have a covariance matrix
    ``noise_covariance`` defined as

    .. math::
        \texttt{noise\_covariance} := \mathbb{E} \bigg (
        \begin{bmatrix}
            v \\ w
        \end{bmatrix}
        \begin{bmatrix}
            v \\ w
        \end{bmatrix}^\mathrm{T}
        \bigg )

    [1] Verhaegen, Michel, and Vincent Verdult. *Filtering and system identification: a least squares approach.*
    Cambridge university press, 2007.
    """
    output_label = 'output'
    """ Label given to an output column in ``self.to_dataframe``. """
    standard_deviation_label = 'standard deviation'
    """ Label given to a standard deviation column in ``self.to_dataframe``. """
    actual_label = 'actual'
    """ Label given to a column in ``self.to_dataframe``, indicating measured values. """
    filtered_label = 'filtered'
    """ Label given to a column in ``self.to_dataframe``, indicating the filtered state of the Kalman filter. """
    next_predicted_label = 'next predicted (no input)'
    """
    Label given to a column in ``self.to_dataframe``, indicating the predicted state of the Kalman filter under the
    absence of further inputs.
    """
    next_predicted_corrected_label = 'next predicted (input corrected)'
    """
    Label given to a column in ``self.to_dataframe``, indicating the predicted state of the Kalman filter corrected
    by previous inputs. The inputs to the state-space model are known, but not at the time that the prediction was
    made. In order to make a fair comparison for prediction performance, the direct effect of the input on the output
    by the matrix :math:`D` is removed in this column.
    
    The latest prediction will have ``np.nan`` in this column, since the input is not yet known.
    """

    def __init__(
            self,
            state_space: StateSpace,
            noise_covariance: np.ndarray
    ):
        self.state_space = state_space

        Utils.validate_matrix_shape(
            noise_covariance,
            (self.state_space.y_dim + self.state_space.x_dim,
             self.state_space.y_dim + self.state_space.x_dim),
            'noise_covariance')
        self.r = noise_covariance[:self.state_space.y_dim, :self.state_space.y_dim]
        self.s = noise_covariance[self.state_space.y_dim:, :self.state_space.y_dim]
        self.q = noise_covariance[self.state_space.y_dim:, self.state_space.y_dim:]

        self.x_filtereds = []
        self.x_predicteds = []
        self.p_filtereds = []
        self.p_predicteds = []
        self.us = []
        self.ys = []
        self.y_filtereds = []
        self.y_predicteds = []
        self.kalman_gains = []

    def step(
            self,
            y: np.ndarray,
            u: np.ndarray
    ):
        """
        Given an observed input ``u`` and output ``y``, update the filtered and predicted states of the Kalman filter.
        Follows the implementation of the conventional Kalman filter in [1] on page 140.

        [1] Verhaegen, Michel, and Vincent Verdult. *Filtering and system identification: a least squares approach.*
        Cambridge university press, 2007.
        """
        Utils.validate_matrix_shape(y, (self.state_space.y_dim, 1), 'y')
        Utils.validate_matrix_shape(u, (self.state_space.u_dim, 1), 'u')

        x_pred = self.x_predicteds[-1] if self.x_predicteds else np.zeros((self.state_space.x_dim, 1))
        p_pred = self.p_predicteds[-1] if self.p_predicteds else np.eye(self.state_space.x_dim)

        k_filtered = p_pred @ self.state_space.c.T @ np.linalg.pinv(
            self.r + self.state_space.c @ p_pred @ self.state_space.c.T
        )
        self.p_filtereds.append(
            p_pred - k_filtered @ self.state_space.c @ p_pred
        )
        self.x_filtereds.append(
            x_pred + k_filtered @ (y - self.state_space.d @ u - self.state_space.c @ x_pred)
        )

        k_pred = (self.s + self.state_space.a @ p_pred @ self.state_space.c.T) @ np.linalg.pinv(
            self.r + self.state_space.c @ p_pred @ self.state_space.c.T
        )
        self.p_predicteds.append(
            self.state_space.a @ p_pred @ self.state_space.a.T
            + self.q
            - k_pred @ (self.s + self.state_space.a @ p_pred @ self.state_space.c.T).T
        )
        self.x_predicteds.append(
            self.state_space.a @ x_pred
            + self.state_space.b @ u
            + k_pred @ (y - self.state_space.d @ u - self.state_space.c @ x_pred)
        )

        self.us.append(u)
        self.ys.append(y)
        self.y_filtereds.append(self.state_space.output(self.x_filtereds[-1], self.us[-1]))
        self.y_predicteds.append(self.state_space.output(self.x_predicteds[-1]))
        self.kalman_gains.append(k_pred)

        return self.y_filtereds[-1], self.y_predicteds[-1]

    def extrapolate(
            self,
            timesteps
    ) -> pd.DataFrame:
        """
        Make a ``timesteps`` number of steps ahead prediction about the output of the state-space model
        ``self.state_space`` given no further inputs.
        The result is a ``pd.DataFrame`` where the columns are ``self.state_space.y_column_names``:
        the output columns of the state-space model ``self.state_space``.
        """
        if not self.x_predicteds:
            raise Exception('Prediction is only possible once Kalman estimation has been performed.')

        state_space = StateSpace(
            self.state_space.a,
            self.state_space.b,
            self.state_space.c,
            self.state_space.d,
            x_init=self.x_predicteds[-1],
            y_column_names=self.state_space.y_column_names,
            u_column_names=self.state_space.u_column_names
        )

        for _ in range(timesteps):
            state_space.step()

        return state_space.to_dataframe()[state_space.y_column_names]

    def _measurement_and_state_standard_deviation(
            self,
            state_covariance_matrices: List[np.ndarray]
    ) -> List[np.ndarray]:
        """
        Calculates the expected standard deviations on the output, assuming independence (!) in the noise of the state
        estimate and the process noise.
        Returns a list of row-vectors containing the standard deviations for the outputs.
        """
        covars_process_y = [
            self.state_space.c @ p @ self.state_space.c.T for p in state_covariance_matrices
        ]

        var_process_ys = [
            np.maximum(
                np.diagonal(p), 0
            )
            for p in covars_process_y
        ]
        var_measurement_y = np.maximum(np.diagonal(self.r), 0)

        return [
            np.sqrt(
                var_process_y + var_measurement_y
            ).reshape(
                (self.state_space.y_dim, 1)
            )
            for var_process_y in var_process_ys
        ]

    @staticmethod
    def _list_of_states_to_array(
            list_of_states: List[np.ndarray]
    ) -> np.ndarray:
        return np.array(list_of_states).squeeze(axis=2)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Returns the output of the Kalman filter as a ``pd.DataFrame``. The returned value contains information about
        filtered and predicted states of the Kalman filter at different timesteps.
        The expected standard deviation of the output is given, assuming independence (!) of the state estimation error
        and measurement noise.

        The rows of the returned dataframe correspond to timesteps.
        The columns of the returned dataframe are a 3-dimensional multi-index with the following levels:

        1. The output name, in the list ``self.state_space.y_column_names``.
        2. An indication of whether the value is
            - a value that was actually measured, these values were given to `self.step` as the `y` parameter,
            - a filtered state,
            - a predicted state given no further input or
            - a predicted state where the effect of the next input has been corrected for.
              This column is useful for comparing prediction performance.
        3. Whether the column is a value or the corresponding expected standard deviation.
        """
        input_corrected_predictions = [
            output + self.state_space.d @ input_state
            for input_state, output
            in zip(self.us[1:], self.y_predicteds[:-1])
        ] + [np.empty((self.state_space.y_dim, 1)) * np.nan]

        output_frames = [
            pd.DataFrame({
                (self.actual_label, self.output_label): outputs,
                (self.filtered_label, self.output_label): filtereds,
                (self.filtered_label, self.standard_deviation_label): filtered_stds,
                (self.next_predicted_label, self.output_label): predicteds,
                (self.next_predicted_label, self.standard_deviation_label): predicted_stds,
                (self.next_predicted_corrected_label, self.output_label): input_corrected_prediction,
                (self.next_predicted_corrected_label, self.standard_deviation_label): predicted_stds
            }).applymap(lambda array: array[0])
            for (
                outputs,
                filtereds,
                predicteds,
                filtered_stds,
                predicted_stds,
                input_corrected_prediction
            ) in zip(
                zip(*self.ys),
                zip(*self.y_filtereds),
                zip(*self.y_predicteds),
                zip(*self._measurement_and_state_standard_deviation(self.p_filtereds)),
                zip(*self._measurement_and_state_standard_deviation(self.p_predicteds)),
                zip(*input_corrected_predictions)
            )
        ]
        return pd.concat(output_frames, axis=1, keys=self.state_space.y_column_names)

    def plot_filtered(self, fig: matplotlib_figure.Figure):  # pragma: no cover
        """
        The top graph plots the filtered output states of the Kalman filter and compares with the measured values.
        The error bars correspond to the expected standard deviations.
        The bottom graph zooms in on the errors between the filtered states and the measured values, compared with
        the expected standard deviations.
        """
        ax1 = fig.add_subplot(2, 1, 1)
        ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)

        df = self.to_dataframe()

        top_legends, bottom_legends = [], []
        for output_name in self.state_space.y_column_names:
            actual_outputs = df[(output_name, self.actual_label, self.output_label)]
            predicted_outputs = df[(output_name, self.filtered_label, self.output_label)]
            std = df[(output_name, self.filtered_label, self.standard_deviation_label)]

            markers, = ax1.plot(
                list(range(len(actual_outputs))),
                actual_outputs,
                'x'
            )
            line, = ax1.plot(
                list(range(len(actual_outputs))),
                actual_outputs,
                '-',
                color=markers.get_color(),
                alpha=.15
            )
            top_legends.append(((markers, line), output_name))

            prediction_errorbar = ax1.errorbar(
                list(range(len(predicted_outputs))),
                predicted_outputs,
                yerr=std,
                marker='_',
                alpha=.5,
                color=markers.get_color(),
                markersize=10,
                linestyle='',
                capsize=3
            )
            top_legends.append((prediction_errorbar, f'Filtered {output_name}'))

            errors = actual_outputs - predicted_outputs
            markers_bottom, = ax2.plot(
                list(range(len(errors))),
                errors,
                'x'
            )
            lines_bottom, = ax2.plot(
                list(range(len(errors))),
                errors,
                '-',
                color=markers_bottom.get_color(),
                alpha=.15
            )
            bottom_legends.append(((markers_bottom, lines_bottom), f'Error {output_name}'))
            prediction_errorbar_bottom, = ax2.plot(
                list(range(len(predicted_outputs))),
                std,
                '--',
                alpha=.5,
                color=markers.get_color(),
            )
            ax2.plot(
                list(range(len(predicted_outputs))),
                -std,
                '--',
                alpha=.5,
                color=markers.get_color(),
            )
            bottom_legends.append((prediction_errorbar_bottom, rf'Filtered $\sigma(${output_name}$)$'))

        lines, names = zip(*top_legends)
        ax1.legend(lines, names, loc='upper left')
        ax1.set_ylabel('Output $y$ (a.u.)')
        ax1.grid()

        lines, names = zip(*bottom_legends)
        ax2.legend(lines, names, loc='upper left')
        ax2.set_xlabel('Index')
        ax2.set_ylabel(r'Filtering error $y-y_{\mathrm{filtered}}$ (a.u.)')
        ax2.grid()
        ax1.set_title('Kalman filter, filtered state')
        plt.setp(ax1.get_xticklabels(), visible=False)

    def plot_predicted(
            self,
            fig: matplotlib_figure.Figure,
            steps_to_extrapolate: int = 1
    ):  # pragma: no cover
        """
        The top graph plots the predicted output states of the Kalman filter and compares with the measured values.
        The error bars correspond to the expected standard deviations.

        The stars on the top right represent the ``steps_to_extrapolate``-steps ahead extrapolation under no further
        inputs. The bottom graph zooms in on the errors between the predicted states and the measured values, compared
        with the expected standard deviations.
        """
        ax1 = fig.add_subplot(2, 1, 1)
        ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)

        df = self.to_dataframe()

        extrapolation = self.extrapolate(steps_to_extrapolate)

        top_legends, bottom_legends = [], []
        for output_name in self.state_space.y_column_names:
            actual_outputs = df[(output_name, self.actual_label, self.output_label)]
            predicted_outputs = df[(output_name, self.next_predicted_corrected_label, self.output_label)]
            std = df[(output_name, self.next_predicted_label, self.standard_deviation_label)]
            last_predicted_std = std.iloc[-1]

            markers, = ax1.plot(
                list(range(len(actual_outputs))),
                actual_outputs,
                'x'
            )
            line, = ax1.plot(
                list(range(len(actual_outputs))),
                actual_outputs,
                '-',
                color=markers.get_color(),
                alpha=.15
            )
            top_legends.append(((markers, line), output_name))

            prediction_errorbar = ax1.errorbar(
                list(range(1, len(predicted_outputs)+1)),
                predicted_outputs,
                yerr=std,
                marker='_',
                alpha=.5,
                color=markers.get_color(),
                markersize=10,
                linestyle='',
                capsize=3
            )
            top_legends.append((prediction_errorbar, f'Predicted {output_name}'))
            extrapolation_errorbar = ax1.errorbar(
                list(range(len(self.ys), len(self.ys) + steps_to_extrapolate)),
                extrapolation[output_name].to_numpy(),
                yerr=last_predicted_std,
                marker='*',
                markersize=9,
                alpha=.8,
                color=markers.get_color(),
                linestyle='',
                capsize=3
            )
            top_legends.append((extrapolation_errorbar, f'Extrapolation {output_name} (no input)'))

            errors = actual_outputs.to_numpy()[1:] - predicted_outputs.to_numpy()[:-1]
            markers_bottom, = ax2.plot(
                list(range(1, len(errors)+1)),
                errors,
                'x'
            )
            lines_bottom, = ax2.plot(
                list(range(1, len(errors)+1)),
                errors,
                '-',
                color=markers_bottom.get_color(),
                alpha=.15
            )
            bottom_legends.append(((markers_bottom, lines_bottom), f'Error {output_name}'))
            prediction_errorbar_bottom, = ax2.plot(
                list(range(1, len(predicted_outputs))),
                std[:-1],
                '--',
                alpha=.5,
                color=markers.get_color(),
            )
            ax2.plot(
                list(range(1, len(predicted_outputs))),
                -std[:-1],
                '--',
                alpha=.5,
                color=markers.get_color(),
            )
            bottom_legends.append((prediction_errorbar_bottom, rf'Predicted $\sigma(${output_name}$)$'))

        lines, names = zip(*top_legends)
        ax1.legend(lines, names, loc='upper left')
        ax1.set_ylabel('Output $y$ (a.u.)')
        ax1.grid()

        lines, names = zip(*bottom_legends)
        ax2.legend(lines, names, loc='upper left')
        ax2.set_xlabel('Index')
        ax2.set_ylabel(r'Prediction error $y-y_{\mathrm{predicted}}$ (a.u.)')
        ax2.grid()
        ax1.set_title('Kalman filter, predicted state')
        plt.setp(ax1.get_xticklabels(), visible=False)


import unittest

import numpy as np

from nfoursid.utils import Utils


class TestUtils(unittest.TestCase):
    def test_block_hankel_matrix(self):
        matrix = np.array(range(15)).reshape((5, 3))
        hankel = Utils.block_hankel_matrix(matrix, 2)
        desired_result = np.array([
            [0., 3., 6., 9.],
            [1., 4., 7., 10.],
            [2., 5., 8., 11.],
            [3., 6., 9., 12.],
            [4., 7., 10., 13.],
            [5., 8., 11., 14.],
        ])
        self.assertTrue(np.all(np.isclose(desired_result, hankel)))

    def test_eigenvalue_decomposition(self):
        matrix = np.fliplr(np.diag(range(1, 3)))
        decomposition = Utils.eigenvalue_decomposition(matrix)
        self.assertTrue(np.all(np.isclose(
            [[0, -1],
             [-1, 0]],
            decomposition.left_orthogonal
        )))
        self.assertTrue(np.all(np.isclose(
            [2, 1],
            np.diagonal(decomposition.eigenvalues)
        )))
        self.assertTrue(np.all(np.isclose(
            [[-1, 0],
             [0, -1]],
            decomposition.right_orthogonal
        )))

        reduced_decomposition = Utils.reduce_decomposition(decomposition, 1)
        self.assertTrue(np.all(np.isclose(
            [[0], [-1]],
            reduced_decomposition.left_orthogonal
        )))
        self.assertTrue(np.all(np.isclose(
            [[2]],
            reduced_decomposition.eigenvalues
        )))
        self.assertTrue(np.all(np.isclose(
            [[-1, 0]],
            reduced_decomposition.right_orthogonal
        )))

    def test_vectorize(self):
        matrix = np.array([
            [0, 2],
            [1, 3]
        ])
        result = Utils.vectorize(matrix)
        self.assertTrue(np.all(np.isclose(
            np.array([
                [0],
                [1],
                [2],
                [3],
            ]),
            result
        )))

    def test_unvectorize(self):
        matrix = np.array([
            [0],
            [1],
            [2],
            [3],
        ])
        result = Utils.unvectorize(matrix, num_rows=2)
        self.assertTrue(np.all(np.isclose(
            np.array([
                [0, 2],
                [1, 3]
            ]),
            result
        )))

        with self.assertRaises(ValueError):
            Utils.unvectorize(matrix, num_rows=3)

        incompatible_matrix = matrix.T
        with self.assertRaises(ValueError):
            Utils.unvectorize(incompatible_matrix, 1)

    def test_validate_matrix_shape(self):
        with self.assertRaises(ValueError):
            Utils.validate_matrix_shape(
                np.array([[0]]),
                (42),
                'error'
            )



from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from nfoursid.state_space import StateSpace
from nfoursid.utils import Utils, Decomposition


class NFourSID:
    r"""
    Perform subspace identification using N4SID [1].
    The state-space model under consideration, is:

    .. math::
        \begin{cases}
            x_{k+1} &= A x_k + B u_k + K e_k \\
            y_k &= C x_k + D u_k + e_k
        \end{cases}

    Data is provided as a dataframe ``dataframe`` where every row is a measurement.
    The output columns are given by ``output_columns``.
    The input columns are given by ``input_columns``.

    The number of block rows to use in the block Hankel matrices is given by ``num_block_rows``.
    If ``num_block_rows`` is chosen to be too big, the computational complexity will increase.
    If ``num_block_rows`` is chosen to be too small, the order of the system might not be possible to determine
    in the eigenvalue diagram. Moreover, if ``num_block_rows`` is chosen to be too small,
    the assumptions of [2] might not hold.

    [1] Van Overschee, Peter, and Bart De Moor. "N4SID: Subspace algorithms for the identification of combined
    deterministic-stochastic systems." Automatica 30.1 (1994): 75-93.
    """
    def __init__(
            self,
            dataframe: pd.DataFrame,
            output_columns: List[str],
            input_columns: List[str] = None,
            num_block_rows: int = 2
    ):
        self.u_columns = input_columns or []
        self.y_columns = output_columns
        self.num_block_rows = num_block_rows

        self._set_input_output_data(dataframe)
        self._initialize_instance_variables()

    def _initialize_instance_variables(self):
        """ Initialize variables. """
        self.R22, self.R32 = None, None
        self.R32_decomposition = None
        self.x_dim = None

    def _set_input_output_data(
            self,
            dataframe: pd.DataFrame
    ):
        """ Perform data consistency checks and set timeseries data arrays. """
        u_frame = dataframe[self.u_columns]
        if u_frame.isnull().any().any():
            raise ValueError('Input data cannot contain nulls')
        y_frame = dataframe[self.y_columns]
        if y_frame.isnull().any().any():
            raise ValueError('Output data cannot contain nulls')
        self.u_array = u_frame.to_numpy()
        self.y_array = y_frame.to_numpy()
        self.u_dim = self.u_array.shape[1]
        self.y_dim = self.y_array.shape[1]

    def subspace_identification(self):
        """
        Perform subspace identification based on the PO-MOESP method.
        The instrumental variable contains past outputs and past inputs.
        The implementation uses a QR-decomposition for numerical efficiency and is based on page 329 of [1].

        A key result of this function is the eigenvalue decomposition of the :math:`R_{32}` matrix
        ``self.R32_decomposition``, based on which the order of the system should be determined.

        [1] Verhaegen, Michel, and Vincent Verdult. *Filtering and system identification: a least squares approach.*
        Cambridge university press, 2007.
        """
        u_hankel = Utils.block_hankel_matrix(self.u_array, self.num_block_rows)
        y_hankel = Utils.block_hankel_matrix(self.y_array, self.num_block_rows)

        u_past, u_future = u_hankel[:, :-self.num_block_rows], u_hankel[:, self.num_block_rows:]
        y_past, y_future = y_hankel[:, :-self.num_block_rows], y_hankel[:, self.num_block_rows:]
        u_instrumental_y = np.concatenate([u_future, u_past, y_past, y_future])

        q, r = map(lambda matrix: matrix.T, np.linalg.qr(u_instrumental_y.T, mode='reduced'))

        y_rows, u_rows = self.y_dim * self.num_block_rows, self.u_dim * self.num_block_rows
        self.R32 = r[-y_rows:, u_rows:-y_rows]
        self.R22 = r[u_rows:-y_rows, u_rows:-y_rows]
        self.R32_decomposition = Utils.eigenvalue_decomposition(self.R32)

    def system_identification(
            self,
            rank: int = None
    ) -> Tuple[StateSpace, np.ndarray]:
        """
        Identify the system matrices of the state-space model given in the description of ``NFourSID``.
        Moreover, the covariance of the measurement-noise and process-noise will be estimated.
        The order of the returned state-space model has rank ``rank`` by reducing the eigenvalue decomposition.
        The implementation is based on page 333 of [1].

        The return value consists of a tuple containing

        - The identified state-space model containing the estimated matrices :math:`(A, B, C, D)`,
        - and an estimate of the covariance matrix of the measurement-noise :math:`w`
          and process-noise :math:`v`.
          The structure of the covariance matrix corresponds to the parameter ``noise_covariance`` of
          ``subspace_identification.kalman.Kalman``.
          See its documentation for more information.

        ``self.system_identification`` needs the QR-decomposition result of subspace identification
        ``self.R32``, and therefore can only be ran after ``self.subspace_identification``.

        [1] Verhaegen, Michel, and Vincent Verdult. *Filtering and system identification: a least squares approach.*
        Cambridge university press, 2007.
        """
        if self.R32_decomposition is None:
            raise Exception('Perform subspace identification first.')
        if rank is None:
            rank = self.y_dim * self.num_block_rows
        self.x_dim = rank

        observability_decomposition = self._get_observability_matrix_decomposition()

        return self._identify_state_space(observability_decomposition)

    def _identify_state_space(
            self,
            observability_decomposition: Decomposition
    ) -> Tuple[StateSpace, np.ndarray]:
        """
        Approximate the row space of the state sequence of a Kalman filter as per the N4SID scheme.
        Then, solve a least squares problem to identify the system matrices.
        Finally, use the residuals to estimate the noise covariance matrix.
        """
        x = (np.power(observability_decomposition.eigenvalues, .5)
             @ observability_decomposition.right_orthogonal)[:, :-1]
        last_y, last_u = self.y_array[self.num_block_rows:, :].T, self.u_array[self.num_block_rows:, :].T
        x_and_y = np.concatenate([x[:, 1:],
                                  last_y[:, :-1]])
        x_and_u = np.concatenate([x[:, :-1],
                                  last_u[:, :-1]])
        abcd = (np.linalg.pinv(x_and_u @ x_and_u.T) @ x_and_u @ x_and_y.T).T
        residuals = x_and_y - abcd @ x_and_u
        covariance_matrix = residuals @ residuals.T / residuals.shape[1]
        q = covariance_matrix[:self.x_dim, :self.x_dim]
        r = covariance_matrix[self.x_dim:, self.x_dim:]
        s = covariance_matrix[:self.x_dim, self.x_dim:]
        state_space_covariance_matrix = np.concatenate(
            [
                np.concatenate([r, s.T], axis=1),
                np.concatenate([s, q], axis=1)
            ],
            axis=0
        )
        return (
            StateSpace(
                abcd[:self.x_dim, :self.x_dim],
                abcd[:self.x_dim, self.x_dim:],
                abcd[self.x_dim:, :self.x_dim],
                abcd[self.x_dim:, self.x_dim:],
            ),
            (state_space_covariance_matrix + state_space_covariance_matrix.T) / 2
        )

    def _get_observability_matrix_decomposition(self) -> Decomposition:
        """
        Calculate the eigenvalue decomposition of the estimate of the observability matrix as per N4SID.
        """
        u_hankel = Utils.block_hankel_matrix(self.u_array, self.num_block_rows)
        y_hankel = Utils.block_hankel_matrix(self.y_array, self.num_block_rows)
        u_and_y = np.concatenate([u_hankel, y_hankel])
        observability = self.R32 @ np.linalg.pinv(self.R22) @ u_and_y
        observability_decomposition = Utils.reduce_decomposition(
            Utils.eigenvalue_decomposition(observability),
            self.x_dim
        )
        return observability_decomposition

    def plot_eigenvalues(self, ax: plt.axes):  # pragma: no cover
        """
        Plot the eigenvalues of the :math:`R_{32}` matrix, so that the order of the state-space model can be determined.
        Since the :math:`R_{32}` matrix should have been calculated, this function can only be used after
        performing ``self.subspace_identification``.
        """
        if self.R32_decomposition is None:
            raise Exception('Perform subspace identification first.')

        ax.semilogy(np.diagonal(self.R32_decomposition.eigenvalues), 'x')
        ax.set_title('Estimated observability matrix decomposition')
        ax.set_xlabel('Index')
        ax.set_ylabel('Eigenvalue')
        ax.grid()



import unittest

import numpy as np

from nfoursid.state_space import StateSpace


class TestStateSpace(unittest.TestCase):
    def setUp(self) -> None:
        self.a = np.array([[2]])
        self.b = np.array([[3]])
        self.c = np.array([[5]])
        self.d = np.array([[7]])
        self.k = np.array([[11]])

    def test_output(self):
        state = np.array([[1]])
        u = np.array([[2]])
        e = np.array([[3]])

        model = StateSpace(
            self.a,
            self.b,
            self.c,
            self.d,
            self.k
        )

        result = model.output(state)
        self.assertAlmostEqual(5, result[0, 0])

        result = model.output(state, u=u)
        self.assertAlmostEqual(19, result[0, 0])

        result = model.output(state, e=e)
        self.assertAlmostEqual(8, result[0, 0])

    def test_step(self):
        u = np.array([[2]])
        e = np.array([[3]])

        model = StateSpace(
            self.a,
            self.b,
            self.c,
            self.d,
            self.k
        )

        y = model.step()
        self.assertAlmostEqual(0, y[0, 0])
        y = model.step(u)
        self.assertAlmostEqual(14, y[0, 0])
        y = model.step(u=u)
        self.assertAlmostEqual(44, y[0, 0])
        y = model.step(e=e)
        self.assertAlmostEqual(93, y[0, 0])

        result = model.to_dataframe().to_numpy()
        self.assertTrue(np.all(np.isclose(
            np.array([
                [0, 0],
                [2, 14],
                [2, 44],
                [0, 93]
            ]),
            result
        )))

    def test_autonomous_system(self):
        model = StateSpace(
            self.a,
            np.zeros((1, 0)),
            self.c,
            np.zeros((1, 0)),
            k=np.array([[1]])
        )
        e = np.array([[1]])

        with self.assertRaises(ValueError):
            model.step(np.array([[1]]))
        y = model.step()
        self.assertEqual((1, 1), y.shape)
        self.assertAlmostEqual(0, y[0, 0])
        y = model.step(e=e)
        self.assertAlmostEqual(1, y[0, 0])
        y = model.step(e=e)
        self.assertAlmostEqual(6, y[0, 0])
        y = model.step(e=e)
        self.assertAlmostEqual(16, y[0, 0])



import unittest

import numpy as np

from nfoursid.nfoursid import NFourSID
from nfoursid.state_space import StateSpace


class TestNFourSid(unittest.TestCase):
    def test_which_is_actually_regression_test(self):
        n_datapoints = 100
        model = StateSpace(
            np.array([[.5]]),
            np.array([[.6]]),
            np.array([[.7]]),
            np.array([[.8]]),
        )
        np.random.seed(0)
        for _ in range(n_datapoints):
            model.step(np.random.standard_normal((1, 1)))

        nfoursid = NFourSID(
            model.to_dataframe(),
            model.y_column_names,
            input_columns=model.u_column_names,
            num_block_rows=2
        )
        nfoursid.subspace_identification()
        identified_model, covariance_matrix = nfoursid.system_identification(rank=1)

        # matrices `a` and `d` don't have freedom of choice: they should be fitted well
        self.assertTrue(is_slightly_close(.5, identified_model.a))
        self.assertTrue(is_slightly_close(.8, identified_model.d))
        self.assertTrue(np.all(is_slightly_close(0, covariance_matrix)))


def is_slightly_close(matrix, number):
    return np.isclose(matrix, number, rtol=0, atol=1e-3)



from typing import List

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib import figure as matplotlib_figure

from nfoursid.utils import Utils


class StateSpace:
    r"""
    A state-space model defined by the following equations:

    .. math::
        \begin{cases}
            x_{k+1} &= A x_k + B u_k + K e_k \\
            y_k &= C x_k + D u_k + e_k
        \end{cases}

    The shapes of the matrices are checked for consistency and will raise if inconsistent.
    If a matrix does not exist in the model representation, the corresponding ``np.ndarray`` should have dimension
    zero along that axis. See the example below.

    Example
    -------
    An autonomous state-space model has no matrices :math:`B` and :math:`D`.
    An autonomous model with a one-dimensional internal state and output, can be represented as follows:

    >>> model = StateSpace(
    >>>     np.ones((1, 1)),
    >>>     np.ones((1, 0)),
    >>>     np.ones((1, 1)),
    >>>     np.ones((1, 0))
    >>> )

    :param a: matrix :math:`A`
    :param b: matrix :math:`B`
    :param c: matrix :math:`C`
    :param d: matrix :math:`D`
    :param k: matrix :math:`K`, optional
    :param x_init: initial state :math:`x_0` of the model, optional
    :param y_column_names: list of output column names, optional
    :param u_column_names: list of input column names, optional
    """
    def __init__(
            self,
            a: np.ndarray,
            b: np.ndarray,
            c: np.ndarray,
            d: np.ndarray,
            k: np.ndarray = None,
            x_init: np.ndarray = None,
            y_column_names: List[str] = None,
            u_column_names: List[str] = None
    ):
        self._set_dimensions(a, b, c)
        self._set_x_init(x_init)
        self._set_column_names(u_column_names, y_column_names)
        self._set_matrices(a, b, c, d, k)

    def _set_dimensions(
            self,
            a: np.ndarray,
            b: np.ndarray,
            c: np.ndarray
    ):
        """ Determine the dimensions of the internal states, outputs and inputs, based on the matrix shapes. """
        self.x_dim = a.shape[0]
        self.y_dim = c.shape[0]
        self.u_dim = b.shape[1]

        if self.y_dim == 0:
            raise ValueError('The dimension of the output should be at least 1')

    def _set_matrices(
            self,
            a: np.ndarray,
            b: np.ndarray,
            c: np.ndarray,
            d: np.ndarray,
            k: np.ndarray
    ):
        """ Validate if the shapes make sense and set the system matrices. """
        if k is None:
            k = np.zeros((self.x_dim, self.y_dim))
        Utils.validate_matrix_shape(a, (self.x_dim, self.x_dim), 'a')
        Utils.validate_matrix_shape(b, (self.x_dim, self.u_dim), 'b')
        Utils.validate_matrix_shape(c, (self.y_dim, self.x_dim), 'c')
        Utils.validate_matrix_shape(d, (self.y_dim, self.u_dim), 'd')
        Utils.validate_matrix_shape(k, (self.x_dim, self.y_dim), 'k')
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.k = k
        self.xs = []
        self.ys = []
        self.us = []

    def _set_column_names(
            self,
            u_column_names: List[str],
            y_column_names: List[str]
    ):
        """ Set the column names of the input and output. """
        if y_column_names is None:
            y_column_names = [f'$y_{i}$' for i in range(self.y_dim)]
        if u_column_names is None:
            u_column_names = [f'$u_{i}$' for i in range(self.u_dim)]
        if len(y_column_names) != self.y_dim:
            raise ValueError(f'Length of `y_column_names` should be {self.y_dim}, not {len(y_column_names)}')
        if len(u_column_names) != self.u_dim:
            raise ValueError(f'Length of `u_column_names` should be {self.u_dim}, not {len(u_column_names)}')
        self.y_column_names = y_column_names
        self.u_column_names = u_column_names

    def _set_x_init(self, x_init: np.ndarray):
        """ Set the initial state, if it is given. """
        if x_init is None:
            x_init = np.zeros((self.x_dim, 1))
        Utils.validate_matrix_shape(x_init, (self.x_dim, 1), 'x_dim')
        self._x_init = x_init

    def step(
            self,
            u: np.ndarray = None,
            e: np.ndarray = None
    ) -> np.ndarray:
        """
        Calculates the output of the state-space model and returns it.
        Updates the internal state of the model as well.
        The input ``u`` is optional, as is the noise ``e``.
        """
        if u is None:
            u = np.zeros((self.u_dim, 1))
        if e is None:
            e = np.zeros((self.y_dim, 1))

        Utils.validate_matrix_shape(u, (self.u_dim, 1), 'u')
        Utils.validate_matrix_shape(e, (self.y_dim, 1), 'e')

        x = self.xs[-1] if self.xs else self._x_init
        x, y = (
            self.a @ x + self.b @ u + self.k @ e,
            self.output(x, u, e)
        )
        self.us.append(u)
        self.xs.append(x)
        self.ys.append(y)
        return y

    def output(
            self,
            x: np.ndarray,
            u: np.ndarray = None,
            e: np.ndarray = None):
        """
        Calculate the output of the state-space model.
        This function calculates the updated :math:`y_k` of the state-space model in the class description.
        The current state ``x`` is required.
        Providing an input ``u`` is optional.
        Providing a noise term ``e`` to be added is optional as well.
        """
        if u is None:
            u = np.zeros((self.u_dim, 1))
        if e is None:
            e = np.zeros((self.y_dim, 1))

        Utils.validate_matrix_shape(x, (self.x_dim, 1), 'x')
        Utils.validate_matrix_shape(u, (self.u_dim, 1), 'u')
        Utils.validate_matrix_shape(e, (self.y_dim, 1), 'e')

        return self.c @ x + self.d @ u + e

    def plot_input_output(self, fig: matplotlib_figure.Figure):  # pragma: no cover
        """
        Given a matplotlib figure ``fig``, plot the inputs and outputs of the state-space model.
        """
        ax1 = fig.add_subplot(2, 1, 1)
        ax2 = fig.add_subplot(2, 1, 2, sharex=ax1)

        for output_name, outputs in zip(self.y_column_names, np.array(self.ys).squeeze(axis=2).T):
            ax1.plot(outputs, label=output_name, alpha=.6)
        ax1.legend(loc='upper right')
        ax1.set_ylabel('Output $y$ (a.u.)')
        ax1.grid()

        for input_name, inputs in zip(self.u_column_names, np.array(self.us).squeeze(axis=2).T):
            ax2.plot(inputs, label=input_name, alpha=.6)
        ax2.legend(loc='upper right')
        ax2.set_ylabel('Input $u$ (a.u.)')
        ax2.set_xlabel('Index')
        ax2.grid()

        ax1.set_title('Inputs and outputs of state-space model')
        plt.setp(ax1.get_xticklabels(), visible=False)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Return the inputs and outputs of the state-space model as a dataframe, where the columns are the input-
        and output-columns.
        """
        inputs_df = pd.DataFrame(np.array(self.us).squeeze(axis=2), columns=self.u_column_names)
        outputs_df = pd.DataFrame(np.array(self.ys).squeeze(axis=2), columns=self.y_column_names)
        return pd.concat([inputs_df, outputs_df], axis=1)



from collections import namedtuple
from typing import Tuple

import numpy as np

Decomposition = namedtuple('Decomposition', ['left_orthogonal', 'eigenvalues', 'right_orthogonal'])
"""
Eigenvalue decomposition of a matrix ``matrix`` such that ``left_orthogonal @ eigenvalues @ right_orthogonal``
equals ``matrix``.
"""


class Utils:
    @staticmethod
    def validate_matrix_shape(
            matrix: np.ndarray,
            shape: Tuple[float, float],
            name: str
    ):
        """
        Raises if ``matrix`` does not have shape ``shape``. The error message will contain ``name``.
        """
        if matrix.shape != shape:
            raise ValueError(f'Dimensions of `{name}` {matrix.shape} are inconsistent. Expected {shape}.')

    @staticmethod
    def eigenvalue_decomposition(
            matrix: np.ndarray
    ) -> Decomposition:
        """
        Calculate eigenvalue decomposition of ``matrix`` as a ``Decomposition``.
        """
        u, eigenvalues, vh = np.linalg.svd(matrix)
        eigenvalues_mat = np.zeros((u.shape[0], vh.shape[0]))
        np.fill_diagonal(eigenvalues_mat, eigenvalues)
        return Decomposition(u, eigenvalues_mat, vh)

    @staticmethod
    def reduce_decomposition(
            decomposition: Decomposition,
            rank: int
    ) -> Decomposition:
        """
        Reduce an eigenvalue decomposition ``decomposition`` such that only ``rank`` number of biggest eigenvalues
        remain. Returns another ``Decomposition``.
        """
        u, s, vh = decomposition
        return Decomposition(
            u[:, :rank],
            s[:rank, :rank],
            vh[:rank, :]
        )

    @staticmethod
    def block_hankel_matrix(
            matrix: np.ndarray,
            num_block_rows: int
    ) -> np.ndarray:
        """
        Calculate a block Hankel matrix based on input matrix ``matrix`` with ``num_block_rows`` block rows.
        The shape of ``matrix`` is interpreted in row-order, like the structure of a ``pd.DataFrame``:
        the rows are measurements and the columns are data sources.

        The returned block Hankel matrix has a columnar structure. Every column of the returned matrix consists
        of ``num_block_rows`` block rows (measurements). See the examples for details.

        Examples
        --------
        Suppose that the input matrix contains 4 measurements of 2-dimensional data:

        >>> matrix = np.array([
        >>>     [0, 1],
        >>>     [2, 3],
        >>>     [4, 5],
        >>>     [6, 7]
        >>> ])

        If the number of block rows is set to ``num_block_rows=2``, then the block Hankel matrix will be

        >>> np.array([
        >>>     [0, 2, 4],
        >>>     [1, 3, 5],
        >>>     [2, 4, 6],
        >>>     [3, 5, 7]
        >>> ])
        """
        hankel_rows_dim = num_block_rows * matrix.shape[1]
        hankel_cols_dim = matrix.shape[0] - num_block_rows + 1

        hankel = np.zeros((hankel_rows_dim, hankel_cols_dim))
        for block_row_index in range(hankel_cols_dim):
            flattened_block_rows = matrix[block_row_index:block_row_index+num_block_rows,
                                          :].flatten()
            hankel[:, block_row_index] = flattened_block_rows
        return hankel

    @staticmethod
    def vectorize(
            matrix: np.ndarray
    ) -> np.ndarray:
        """
        Given a matrix ``matrix`` of shape ``(a, b)``, return a vector of shape ``(a*b, 1)`` with all columns of
        ``matrix`` stacked on top of eachother.
        """
        return np.reshape(matrix.flatten(order='F'), (matrix.shape[0] * matrix.shape[1], 1))

    @staticmethod
    def unvectorize(
            vector: np.ndarray,
            num_rows: int
    ) -> np.ndarray:
        """
        Given a vector ``vector`` of shape ``(num_rows*b, 1)``, return a matrix of shape ``(num_rows, b)`` such that
        the stacked columns of the returned matrix equal ``vector``.
        """
        if vector.shape[0] % num_rows != 0 or vector.shape[1] != 1:
            raise ValueError(f'Vector shape {vector.shape} and `num_rows`={num_rows} are incompatible')
        return vector.reshape((num_rows, vector.shape[0] // num_rows), order='F')



import copy
import unittest

import numpy as np

from nfoursid.kalman import Kalman
from nfoursid.state_space import StateSpace


class TestKalman(unittest.TestCase):
    def setUp(self) -> None:
        self.x_init = np.array([[2]])
        self.model = StateSpace(
            np.array([[.5]]),
            np.array([[.6]]),
            np.array([[.7]]),
            np.array([[.8]]),
            np.array([[1]]),
            x_init=self.x_init
        )

    def test_step(self):
        n_datapoints = 100
        noise_reduction = 1e5

        kalman = Kalman(copy.deepcopy(self.model), np.eye(2) / (noise_reduction ** 2))
        np.random.seed(0)
        for _ in range(n_datapoints):
            u = np.random.standard_normal((1, 1))

            y = self.model.step(u, np.random.standard_normal((1, 1)) / noise_reduction)
            kalman.step(y, u)

        self.assertTrue(is_slightly_close(
            2.0593793476904603,
            kalman.y_filtereds[-1]
        ))

    def test_extrapolate(self):
        n_to_predict = 4
        kalman = Kalman(copy.deepcopy(self.model), np.eye(2))
        kalman.x_predicteds = [
            self.x_init
        ]
        predictions = kalman.extrapolate(n_to_predict)

        self.assertTrue(np.all(np.isclose(
            np.array([
                [1.4],
                [.7],
                [.35],
                [.175]
            ]),
            predictions.to_numpy()
        )))

    def test_measurement_and_state_standard_deviation(self):
        n_steps = 3
        noise_variance = 9
        state_variance = 16 / self.model.c[0, 0] ** 2

        kalman = Kalman(self.model, noise_variance * np.eye(2))
        state_covariance_matrices = [
            state_variance * np.eye(1) for _ in range(n_steps)
        ]

        output_standard_deviations = kalman._measurement_and_state_standard_deviation(state_covariance_matrices)

        self.assertEqual(n_steps, len(output_standard_deviations))
        for output_standard_deviation in output_standard_deviations:
            self.assertTrue(np.isclose(5, output_standard_deviation))

    def test_list_of_states_to_array(self):
        list_of_states = [
            np.array([[1], [2]]),
            np.array([[3], [4]])
        ]
        result = Kalman._list_of_states_to_array(list_of_states)
        self.assertTrue(np.all(np.isclose(
            np.array([
                [1, 2],
                [3, 4]
            ]),
            result
        )))

    def test_to_dataframe(self):
        kalman = Kalman(self.model, np.eye(2))
        kalman.us = [
            np.array([[1]]),
            np.array([[1]]),
        ]
        kalman.ys = [
            np.array([[1]]),
            np.array([[3]]),
        ]
        kalman.y_filtereds = [
            np.array([[11]]),
            np.array([[13]]),
        ]
        kalman.y_predicteds = [
            np.array([[21]]),
            np.array([[23]]),
        ]
        kalman.p_filtereds = 2 * [np.eye(1)]
        kalman.p_predicteds = 2 * [np.eye(1)]
        df = kalman.to_dataframe()

        self.assertTrue(np.all(np.isclose(
            np.array([1, 3]),
            df[('$y_0$', kalman.actual_label, kalman.output_label)].to_numpy()
        )))
        self.assertTrue(np.all(np.isclose(
            np.array([11, 13]),
            df[('$y_0$', kalman.filtered_label, kalman.output_label)].to_numpy()
        )))
        self.assertTrue(np.all(np.isclose(
            np.array([21, 23]),
            df[('$y_0$', kalman.next_predicted_label, kalman.output_label)].to_numpy()
        )))
        self.assertTrue(np.all(np.isclose(
            np.array([21.8]),
            df[('$y_0$', kalman.next_predicted_corrected_label, kalman.output_label)].to_numpy()[:-1]
        )))
        self.assertTrue(np.isnan(
            df[('$y_0$', kalman.next_predicted_corrected_label, kalman.output_label)].iloc[-1]
        ))


def is_slightly_close(matrix, number):
    return np.isclose(matrix, number, rtol=0, atol=1e-3)
